import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, cast

import sqlalchemy

from edgy.core.connection.database import Database
from edgy.core.db.context_vars import set_schema
from edgy.types import Undefined

if TYPE_CHECKING:
    from edgy import QuerySet


class QuerySetPropsMixin:
    """
    Properties used by the Queryset are placed in isolation
    for clean access and maintainance.
    """

    _table: Optional["sqlalchemy.Table"] = None
    _database: Optional["Database"] = None

    @property
    def database(self) -> Database:
        if self._database is None:
            return cast("Database", self.model_class.database)
        return self._database

    @database.setter
    def database(self, value: Database) -> None:
        self._database = value

    @property
    def table(self) -> sqlalchemy.Table:
        if self._table is None:
            return cast("sqlalchemy.Table", self.model_class.table_schema(self.active_schema))
        return self._table

    @table.setter
    def table(self, value: Optional[sqlalchemy.Table]) -> None:
        self._table = value

    @property
    def pknames(self) -> Sequence[str]:
        return self.model_class.pknames  # type: ignore

    @property
    def pkcolumns(self) -> Sequence[str]:
        return self.model_class.pkcolumns  # type: ignore


_sentinel = object()


class TenancyMixin:
    """
    Mixin used for querying a possible multi tenancy and multi-database application
    """

    def using(
        self,
        _positional: Any = _sentinel,
        *,
        database: Union[str, Any, None, "Database"] = Undefined,
        schema: Union[str, Any, None, Literal[False]] = Undefined,
    ) -> "QuerySet":
        """
        Enables and switches the db schema and/or db.

        Generates the registry object pointing to the desired schema
        using the same connection.

        Use schema=False to unset the schema to the active default schema.
        Use database=None to use the default database again.
        """
        if _positional is not _sentinel:
            warnings.warn(
                "Passing positional arguments to using is deprecated. Use schema= instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            schema = _positional
            if schema is Undefined:
                schema = False

        queryset = cast("QuerySet", self._clone())
        if database is not Undefined:
            if isinstance(database, Database):
                connection: Database = database
            elif database is None:
                connection = self.model_class.meta.registry.database
            else:
                assert (
                    database is None or database in self.model_class.meta.registry.extra
                ), f"`{database}` is not in the connections extra of the model`{self.model_class.__name__}` registry"

                connection = self.model_class.meta.registry.extra[database]
            queryset.database = connection
        if schema is not Undefined:
            queryset.using_schema = schema if schema is not False else Undefined
            queryset.active_schema = queryset.get_schema()
            queryset.table = None
        return queryset

    def using_with_db(
        self, connection_name: str, schema: Union[str, Any, None, Literal[False]] = Undefined
    ) -> "QuerySet":
        """
        Switches the database connection and schema
        """
        warnings.warn(
            "'using_with_db' is deprecated in favor of 'using' with schema, database arguments.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.using(database=connection_name, schema=schema)


def activate_schema(tenant_name: str) -> None:
    """
    Activates the tenant for the context of the query.
    """
    set_schema(tenant_name)


def deactivate_schema() -> None:
    """
    Deactivates the tenant for the context of the query.
    """
    set_schema(None)



import inspect
from typing import Any

from sqlalchemy.orm import Mapped, relationship


class DeclarativeMixin:
    """
    Mixin for declarative base models.
    """

    @classmethod
    def declarative(cls) -> Any:
        return cls.generate_model_declarative()

    @classmethod
    def generate_model_declarative(cls) -> Any:
        """
        Transforms a core Edgy table into a Declarative model table.
        """
        Base = cls.meta.registry.declarative_base

        # Build the original table
        fields = {"__table__": cls.table}

        # Generate base
        model_table = type(cls.__name__, (Base,), fields)

        # Make sure if there are foreignkeys, builds the relationships
        for column in cls.table.columns:
            if not column.foreign_keys:
                continue

            # Maps the relationships with the foreign keys and related names
            field = cls.meta.fields.get(column.name)
            to = field.to.__name__ if inspect.isclass(field.to) else field.to
            mapped_model: Mapped[to] = relationship(to)  # type: ignore

            # Adds to the current model
            model_table.__mapper__.add_property(f"{column.name}_relation", mapped_model)

        return model_table



import os
from functools import cached_property
from typing import Any, Optional

from pydantic_settings import SettingsConfigDict

from edgy.conf.global_settings import EdgySettings


class TenancySettings(EdgySettings):
    """
    BaseSettings used for the contrib of Edgy tenancy
    """

    model_config = SettingsConfigDict(extra="allow", ignored_types=(cached_property,))
    auto_create_schema: bool = True
    auto_drop_schema: bool = False
    tenant_schema_default: str = "public"
    tenant_model: Optional[str] = None
    domain: Any = os.getenv("DOMAIN")
    domain_name: str = "localhost"
    auth_user_model: Optional[str] = None



from .base import BaseField
from .composite_field import CompositeField
from .core import (
    AutoNowMixin,
    BigIntegerField,
    BinaryField,
    BooleanField,
    CharField,
    ChoiceField,
    DateField,
    DateTimeField,
    DecimalField,
    EmailField,
    FloatField,
    IntegerField,
    IPAddressField,
    JSONField,
    PasswordField,
    SmallIntegerField,
    TextField,
    TimeField,
    URLField,
    UUIDField,
)
from .exclude_field import ExcludeField
from .file_field import FileField
from .foreign_keys import ForeignKey
from .image_field import ImageField
from .many_to_many import ManyToMany, ManyToManyField
from .one_to_one_keys import OneToOne, OneToOneField
from .ref_foreign_key import RefForeignKey
from .types import BaseFieldType

__all__ = [
    "AutoNowMixin",
    "BaseField",
    "BaseFieldType",
    "BigIntegerField",
    "BinaryField",
    "BooleanField",
    "CharField",
    "ChoiceField",
    "CompositeField",
    "DateField",
    "DateTimeField",
    "DecimalField",
    "EmailField",
    "ExcludeField",
    "FloatField",
    "FileField",
    "ImageField",
    "ForeignKey",
    "IntegerField",
    "ImageFile",
    "IPAddressField",
    "JSONField",
    "RefForeignKey",
    "ManyToMany",
    "ManyToManyField",
    "OneToOne",
    "OneToOneField",
    "PasswordField",
    "SmallIntegerField",
    "TextField",
    "TimeField",
    "URLField",
    "UUIDField",
]



import datetime
import decimal
import enum
import ipaddress
import uuid
from collections.abc import Callable, Sequence
from enum import EnumMeta
from functools import partial
from re import Pattern
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import pydantic
import sqlalchemy
from pydantic import EmailStr

from edgy.core.db.context_vars import CURRENT_INSTANCE, CURRENT_PHASE, EXPLICIT_SPECIFIED_VALUES
from edgy.core.db.fields._internal import IPAddress
from edgy.core.db.fields._validators import IPV4_REGEX, IPV6_REGEX
from edgy.core.db.fields.base import Field
from edgy.core.db.fields.factories import FieldFactory
from edgy.core.db.fields.types import BaseFieldType
from edgy.exceptions import FieldDefinitionError

if TYPE_CHECKING:
    import zoneinfo

    from edgy.core.db.models.types import BaseModelType


CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]


class CharField(FieldFactory, str):
    """String field representation that constructs the Field class and populates the values"""

    field_type = str

    def __new__(  # type: ignore
        cls,
        *,
        min_length: Optional[int] = None,
        regex: Union[str, Pattern] = None,
        pattern: Union[str, Pattern] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        if pattern is None:
            pattern = regex
        del regex
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        max_length = kwargs.get("max_length", 0)
        if max_length is not None and max_length <= 0:
            raise FieldDefinitionError(detail=f"'max_length' is required for {cls.__name__}")

        min_length = kwargs.get("min_length")
        pattern = kwargs.get("regex")

        assert min_length is None or isinstance(min_length, int)
        assert max_length is None or isinstance(max_length, int)
        assert pattern is None or isinstance(pattern, (str, Pattern))

    @classmethod
    def get_column_type(cls, max_length: Optional[int] = None, **kwargs: Any) -> Any:
        return (
            sqlalchemy.Text(collation=kwargs.get("collation"))
            if max_length is None
            else sqlalchemy.String(length=max_length, collation=kwargs.get("collation"))
        )


class TextField(CharField):
    """String representation of a text field which means no max_length required"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", None)
        super().validate(kwargs)


class IncrementOnSaveBaseField(Field):
    increment_on_save: int = 0

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            **kwargs,
        )
        if self.increment_on_save != 0:
            self.pre_save_callback = self._notset_pre_save_callback

    async def _notset_pre_save_callback(
        self, value: Any, original_value: Any, force_insert: bool, instance: "BaseModelType"
    ) -> dict[str, Any]:
        explicit_values = EXPLICIT_SPECIFIED_VALUES.get()
        if explicit_values is not None and self.name in explicit_values:
            return {}
        model_or_query = CURRENT_INSTANCE.get()

        if force_insert:
            if original_value is None:
                return {self.name: self.get_default_value()}
            else:
                return {self.name: value + self.increment_on_save}
        elif not self.primary_key:
            # update path
            return {
                self.name: (
                    model_or_query if model_or_query is not None else instance
                ).table.columns[self.name]
                + self.increment_on_save
            }
        else:
            # update path
            return {}

    def get_default_values(
        self,
        field_name: str,
        cleaned_data: dict[str, Any],
    ) -> dict[str, Any]:
        if self.increment_on_save != 0:
            phase = CURRENT_PHASE.get()
            if phase in "prepare_update":
                return {field_name: None}
        return super().get_default_values(field_name, cleaned_data)

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        phase = CURRENT_PHASE.get()
        instance = CURRENT_INSTANCE.get()
        if self.increment_on_save != 0 and not self.primary_key and phase == "post_update":
            # a bit dirty but works
            instance.__dict__.pop(field_name, None)
            return {}
        return super().to_model(field_name, value)


class IntegerField(FieldFactory, int):
    """
    Integer field factory that construct Field classes and populated their values.
    """

    field_type = int
    field_bases = (IncrementOnSaveBaseField,)

    def __new__(  # type: ignore
        cls,
        *,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        multiple_of: Optional[int] = None,
        increment_on_save: int = 0,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Integer()

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        increment_on_save = kwargs.get("increment_on_save", 0)
        if increment_on_save == 0 and kwargs.get("primary_key", False):
            kwargs.setdefault("autoincrement", True)
        if increment_on_save != 0:
            if kwargs.get("autoincrement"):
                raise FieldDefinitionError(
                    detail="'autoincrement' is incompatible with 'increment_on_save'"
                )
            if kwargs.get("null"):
                raise FieldDefinitionError(
                    detail="'null' is incompatible with 'increment_on_save'"
                )
            kwargs.setdefault("read_only", True)
            kwargs["inject_default_on_partial_update"] = True


class FloatField(FieldFactory, float):
    """Representation of a int32 and int64"""

    field_type = float

    def __new__(  # type: ignore
        cls,
        *,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Float(asdecimal=False)


class BigIntegerField(IntegerField):
    """Representation of big integer field"""

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        if kwargs.get("autoincrement", False):
            kwargs.setdefault("skip_reflection_type_check", True)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        # sqlite special we cannot have a big IntegerField as PK
        if kwargs.get("autoincrement"):
            return sqlalchemy.BigInteger().with_variant(sqlalchemy.Integer, "sqlite")
        return sqlalchemy.BigInteger()


class SmallIntegerField(IntegerField):
    """Represents a small integer field"""

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.SmallInteger()


class DecimalField(FieldFactory, decimal.Decimal):
    field_type = decimal.Decimal

    def __new__(  # type: ignore
        cls,
        *,
        ge: Union[int, float, decimal.Decimal, None] = None,
        gt: Union[int, float, decimal.Decimal, None] = None,
        le: Union[int, float, decimal.Decimal, None] = None,
        lt: Union[int, float, decimal.Decimal, None] = None,
        multiple_of: Union[int, decimal.Decimal, None] = None,
        max_digits: Optional[int] = None,
        decimal_places: Optional[int] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in ["cls", "__class__", "kwargs"]},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Numeric(
            precision=kwargs.get("max_digits"), scale=kwargs.get("decimal_places"), asdecimal=True
        )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)

        decimal_places = kwargs.get("decimal_places")
        if decimal_places is None or decimal_places < 0:
            raise FieldDefinitionError("decimal_places are required for DecimalField")


class BooleanField(FieldFactory, int):
    """Representation of a boolean"""

    field_type = bool

    def __new__(  # type: ignore
        cls,
        *,
        default: Optional[bool] = False,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Boolean()


class TimezonedField:
    default_timezone: Optional["zoneinfo.ZoneInfo"]
    force_timezone: Optional["zoneinfo.ZoneInfo"]
    remove_timezone: bool

    def _convert_datetime(
        self, value: datetime.datetime
    ) -> Union[datetime.datetime, datetime.date]:
        if value.tzinfo is None and self.default_timezone is not None:
            value = value.replace(tzinfo=self.default_timezone)
        if self.force_timezone is not None and value.tzinfo != self.force_timezone:
            if value.tzinfo is None:
                value = value.replace(tzinfo=self.force_timezone)
            else:
                value = value.astimezone(self.force_timezone)
        if self.remove_timezone:
            value = value.replace(tzinfo=None)
        if self.field_type is datetime.date:
            return value.date()
        return value

    def check(self, value: Any) -> Optional[Union[datetime.datetime, datetime.date]]:
        if value is None:
            return None
        elif isinstance(value, datetime.datetime):
            return self._convert_datetime(value)
        elif isinstance(value, (int, float)):
            return self._convert_datetime(
                datetime.datetime.fromtimestamp(value, self.default_timezone)
            )
        elif isinstance(value, str):
            return self._convert_datetime(datetime.datetime.fromisoformat(value))
        elif isinstance(value, datetime.date):
            # datetime is subclass, so check datetime first

            # don't touch dates when DateField
            if self.field_type is datetime.date:
                return value
            return self._convert_datetime(
                datetime.datetime(year=value.year, month=value.month, day=value.day)
            )
        else:
            raise ValueError(f"Invalid type detected: {type(value)}")

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Optional[Union[datetime.datetime, datetime.date]]]:
        """
        Convert input object to datetime
        """
        return {field_name: self.check(value)}

    def get_default_value(self) -> Any:
        return self.check(super().get_default_value())


class AutoNowMixin(FieldFactory):
    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        if auto_now_add and auto_now:
            raise FieldDefinitionError("'auto_now' and 'auto_now_add' cannot be both True")

        if auto_now_add or auto_now:
            kwargs.setdefault("read_only", True)
            kwargs["inject_default_on_partial_update"] = auto_now

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        if auto_now_add or auto_now:
            # date.today cannot handle timezone so use alway datetime and convert back to date
            kwargs["default"] = partial(datetime.datetime.now, default_timezone)
        return super().__new__(cls, **kwargs)


class DateTimeField(AutoNowMixin, datetime.datetime):
    """Representation of a datetime field"""

    field_type = datetime.datetime
    field_bases = (TimezonedField, Field)

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        remove_timezone: bool = False,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        kwargs.setdefault("with_timezone", not remove_timezone)
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, with_timezone: bool = True, **kwargs: Any) -> Any:
        return sqlalchemy.DateTime(with_timezone)

    @classmethod
    def get_default_values(
        cls,
        field_obj: Field,
        field_name: str,
        cleaned_data: dict[str, Any],
        original_fn: Any = None,
    ) -> Any:
        phase = CURRENT_PHASE.get()
        if field_obj.auto_now_add and phase == "prepare_update":
            return {}
        return original_fn(field_name, cleaned_data)


class DateField(AutoNowMixin, datetime.date):
    """Representation of a date field"""

    field_type = datetime.date
    field_bases = (TimezonedField, Field)

    def __new__(  # type: ignore
        cls,
        *,
        auto_now: Optional[bool] = False,
        auto_now_add: Optional[bool] = False,
        default_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        force_timezone: Optional["zoneinfo.ZoneInfo"] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        # the datetimes lose the information anyway
        kwargs["remove_timezone"] = False

        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Date()


class TimeField(FieldFactory, datetime.time):
    """Representation of a time field"""

    field_type = datetime.time

    def __new__(cls, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.Time(kwargs.get("with_timezone") or False)


class JSONField(FieldFactory, pydantic.Json):  # type: ignore
    """Representation of a JSONField"""

    field_type = Any

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.JSON()


class BinaryField(FieldFactory, bytes):
    """Representation of a binary"""

    field_type = bytes

    def __new__(cls, *, max_length: Optional[int] = None, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.LargeBinary(kwargs.get("max_length"))


class UUIDField(FieldFactory, uuid.UUID):
    """Representation of a uuid"""

    field_type = uuid.UUID

    def __new__(cls, **kwargs: Any) -> BaseFieldType:  # type: ignore
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        return sqlalchemy.UUID()


class ChoiceField(FieldFactory):
    """Representation of an Enum"""

    field_type = enum.Enum

    def __new__(  # type: ignore
        cls,
        choices: Optional[Sequence[Union[tuple[str, str], tuple[str, int]]]] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{k: v for k, v in locals().items() if k not in CLASS_DEFAULTS},
        }
        return super().__new__(cls, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        choice_class = kwargs.get("choices")
        if choice_class is None or not isinstance(choice_class, EnumMeta):
            raise FieldDefinitionError("ChoiceField choices must be an Enum")

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> sqlalchemy.Enum:
        choice_class = kwargs.get("choices")
        return sqlalchemy.Enum(choice_class)


class PasswordField(CharField):
    """
    Representation of a Password
    """

    def __new__(  # type: ignore
        cls,
        derive_fn: Optional[Callable[[str], str]] = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs.setdefault("keep_original", derive_fn is not None)
        return super().__new__(cls, derive_fn=derive_fn, **kwargs)

    @classmethod
    def to_model(
        cls, field_obj: BaseFieldType, field_name: str, value: Any, original_fn: Any = None
    ) -> dict[str, Any]:
        if isinstance(value, (tuple, list)):
            if value[0] != value[1]:
                raise ValueError("Password doesn't match.")
            else:
                value = value[0]
        retval: dict[str, Any] = {}
        phase = CURRENT_PHASE.get()
        derive_fn = cast(Optional[Callable[[str], str]], field_obj.derive_fn)
        if phase in {"set", "init"} and derive_fn is not None:
            retval[field_name] = derive_fn(value)
            if getattr(field_obj, "keep_original", False):
                retval[f"{field_name}_original"] = value
        else:
            retval[field_name] = value
            # blank after saving or loading
            if phase in {"post_insert", "post_update", "load"} and getattr(
                field_obj, "keep_original", False
            ):
                retval[f"{field_name}_original"] = None

        return retval

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("secret", True)
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class EmailField(CharField):
    field_type = EmailStr

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class URLField(CharField):
    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        kwargs.setdefault("max_length", 255)
        super().validate(kwargs)


class IPAddressField(FieldFactory, str):
    field_type = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

    def __new__(  # type: ignore
        cls,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }

        return super().__new__(cls, **kwargs)

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> IPAddress:
        return IPAddress()

    @staticmethod
    def is_native_type(value: str) -> bool:
        return isinstance(value, (ipaddress.IPv4Address, ipaddress.IPv6Address))

    # overwrite
    @classmethod
    def check(cls, field_obj: BaseFieldType, value: Any, original_fn: Any = None) -> Any:
        if cls.is_native_type(value):
            return value

        match_ipv4 = IPV4_REGEX.match(value)
        match_ipv6 = IPV6_REGEX.match(value)

        if not match_ipv4 and not match_ipv6:  # type: ignore
            raise ValueError("Must be a valid IP format.")

        try:
            return ipaddress.ip_address(value)
        except ValueError:
            raise ValueError("Must be a real IP.")  # noqa



import asyncio
import logging
import os
import sys
from logging.config import fileConfig
from typing import Any, Optional

from alembic import context
from rich.console import Console

from edgy.cli.constants import APP_PARAMETER, EDGY_DB
from edgy.cli.env import MigrationEnv
from edgy.core.connection import Database
from edgy.exceptions import EdgyException

# The console used for the outputs
console = Console()

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config: Any = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)
logger = logging.getLogger("alembic.env")


def get_app_location(argv: Any) -> Any:
    """
    Manually checks for the --app parameter.
    """
    if APP_PARAMETER in argv:
        try:
            return argv[argv.index(APP_PARAMETER) + 1]
        except IndexError as e:
            raise EdgyException(detail=str(e))  # noqa
    return None


def get_app() -> Any:
    """
    Gets the app via environment variable or via console parameter.
    """
    app_path = get_app_location(sys.argv[1:])
    migration = MigrationEnv()
    app_env = migration.load_from_env(path=app_path, enable_logging=False)
    return app_env.app


def get_engine_url() -> str:
    url: Optional[str] = os.environ.get("EDGY_DATABASE_URL")
    assert url, '"EDGY_DATABASE_URL" not specified or empty'
    return url


app: Any = get_app()

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
config.set_main_option("sqlalchemy.url", get_engine_url())

target_db = getattr(app, EDGY_DB)["migrate"].registry

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_metadata() -> Any:
    if hasattr(target_db, "metadatas"):
        return target_db.metadatas[None]
    return target_db.metadata


def run_migrations_offline() -> Any:
    """
    Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well. By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.
    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=get_metadata(), literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Any) -> Any:
    # this callback is used to prevent an auto-migration from being generated
    # when there are no changes to the schema
    # reference: http://alembic.zzzcomputing.com/en/latest/cookbook.html
    def process_revision_directives(context, revision, directives) -> Any:  # type: ignore
        if getattr(config.cmd_opts, "autogenerate", False):
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []
                console.print("[bright_red]No changes in schema detected.")

    context.configure(
        connection=connection,
        target_metadata=get_metadata(),
        process_revision_directives=process_revision_directives,
        **getattr(app, EDGY_DB)["migrate"].kwargs,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_migrations_online() -> Any:
    """
    Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # the original script checked for the async compatibilit
    # we are only compatible with async drivers so just use Database
    async with Database(get_engine_url()) as database:
        await database.run_sync(do_run_migrations)


if context.is_offline_mode():
    run_migrations_offline()
else:
    asyncio.run(run_migrations_online())



from collections.abc import Sequence
from functools import lru_cache, partial
from typing import Any, Literal, Union, cast

from edgy.core.db.constants import CASCADE, RESTRICT, SET_NULL
from edgy.core.db.fields.base import Field
from edgy.core.db.fields.types import BaseFieldType
from edgy.exceptions import FieldDefinitionError

CLASS_DEFAULTS = ["cls", "__class__", "kwargs"]

default_methods_overwritable_by_factory: set[str] = {
    key
    for key, attr in BaseFieldType.__dict__.items()
    if callable(attr) and not key.startswith("_")
}
default_methods_overwritable_by_factory.discard("get_column_names")
default_methods_overwritable_by_factory.discard("__init__")

# extra methods
default_methods_overwritable_by_factory.add("__set__")
default_methods_overwritable_by_factory.add("__get__")
default_methods_overwritable_by_factory.add("modify_input")
default_methods_overwritable_by_factory.add("post_save_callback")
default_methods_overwritable_by_factory.add("post_delete_callback")

# BaseCompositeField
default_methods_overwritable_by_factory.add("translate_name")
default_methods_overwritable_by_factory.add("get_composite_fields")

# Field
default_methods_overwritable_by_factory.add("check")
default_methods_overwritable_by_factory.add("get_column")

# RelationshipField
default_methods_overwritable_by_factory.add("traverse_field")
default_methods_overwritable_by_factory.add("is_cross_db")

# ForeignKey
default_methods_overwritable_by_factory.add("expand_relationship")


class FieldFactoryMeta(type):
    def __instancecheck__(self, instance: Any) -> bool:
        return super().__instancecheck__(instance) or isinstance(
            instance, self._get_field_cls(self)
        )

    def __subclasscheck__(self, subclass: Any) -> bool:
        return super().__subclasscheck__(subclass) or issubclass(
            subclass, self._get_field_cls(self)
        )


class FieldFactory(metaclass=FieldFactoryMeta):
    """The base for all model fields to be used with Edgy"""

    field_bases: Sequence[Any] = (Field,)
    field_type: Any = None
    methods_overwritable_by_factory: frozenset[str] = frozenset(
        default_methods_overwritable_by_factory
    )

    def __new__(cls, **kwargs: Any) -> BaseFieldType:
        cls.validate(kwargs)
        return cls.build_field(**kwargs)

    @classmethod
    def build_field(cls, **kwargs: Any) -> BaseFieldType:
        column_type = cls.get_column_type(**kwargs)
        pydantic_type = cls.get_pydantic_type(**kwargs)
        constraints = cls.get_constraints(**kwargs)

        new_field = cls._get_field_cls(cls)

        new_field_obj: BaseFieldType = new_field(  # type: ignore
            field_type=pydantic_type,
            annotation=pydantic_type,
            column_type=column_type,
            constraints=constraints,
            factory=cls,
            **kwargs,
        )
        cls.overwrite_methods(new_field_obj)
        return new_field_obj

    @classmethod
    def overwrite_methods(cls, field_obj: BaseFieldType) -> None:
        """Called in metaclasses"""
        for key in dir(cls):
            if key in cls.methods_overwritable_by_factory and hasattr(cls, key):
                fn = getattr(cls, key)
                original_fn = getattr(field_obj, key, None)
                # use original func, not the wrapper
                if hasattr(fn, "__func__"):
                    fn = fn.__func__

                # fix classmethod, prevent injection of self or class
                setattr(
                    field_obj,
                    key,
                    # .__func__ is a workaround for python < 3.10, python >=3.10 works without
                    staticmethod(partial(fn, cls, field_obj, original_fn=original_fn)).__func__,
                )

    @classmethod
    def repack(cls, field_obj: BaseFieldType) -> None:
        for key in dir(cls):
            if key in cls.methods_overwritable_by_factory and hasattr(cls, key):
                packed_fn = getattr(field_obj, key, None)
                if packed_fn is not None and hasattr(packed_fn, "func"):
                    setattr(
                        field_obj,
                        key,
                        # .__func__ is a workaround for python < 3.10, python >=3.10 works without
                        staticmethod(
                            partial(packed_fn.func, cls, field_obj, **packed_fn.keywords)
                        ).__func__,
                    )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:  # pragma no cover
        """
        Used to validate if all required parameters on a given field type are set and modify them if needed.
        :param kwargs: dict with all params passed during construction
        :type kwargs: Any
        """

    @classmethod
    def get_constraints(cls, **kwargs: Any) -> Sequence[Any]:
        """Returns the propery column type for the field, None for Metafields"""
        return []

    @classmethod
    def get_column_type(cls, **kwargs: Any) -> Any:
        """Returns the propery column type for the field, None for Metafields"""
        return None

    @classmethod
    def get_pydantic_type(cls, **kwargs: Any) -> Any:
        """Returns the type for pydantic"""
        return cls.field_type

    @staticmethod
    @lru_cache(None)
    def _get_field_cls(cls: "FieldFactory") -> BaseFieldType:
        return cast(BaseFieldType, type(cls.__name__, cast(Any, cls.field_bases), {}))


class ForeignKeyFieldFactory(FieldFactory):
    """The base for all model fields to be used with Edgy"""

    field_type: Any = Any

    def __new__(
        cls,
        *,
        to: Any = None,
        on_update: str = CASCADE,
        on_delete: str = RESTRICT,
        related_name: Union[str, Literal[False]] = "",
        server_onupdate: Any = None,
        **kwargs: Any,
    ) -> BaseFieldType:
        kwargs = {
            **kwargs,
            **{key: value for key, value in locals().items() if key not in CLASS_DEFAULTS},
        }
        cls.validate(kwargs)
        # update related name when available
        return cls.build_field(**kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        """default validation useful for one_to_one and foreign_key"""
        on_update = kwargs.get("on_update", CASCADE)
        on_delete = kwargs.get("on_delete", RESTRICT)
        kwargs.setdefault("null", False)
        null = kwargs["null"]

        if on_delete is None:
            raise FieldDefinitionError("on_delete must not be null.")

        if on_delete == SET_NULL and not null:
            raise FieldDefinitionError("When SET_NULL is enabled, null must be True.")

        if on_update and (on_update == SET_NULL and not null):
            raise FieldDefinitionError("When SET_NULL is enabled, null must be True.")
        related_name = kwargs.get("related_name", "")

        # tolerate None and False
        if related_name and not isinstance(related_name, str):
            raise FieldDefinitionError("related_name must be a string.")

        if related_name:
            kwargs["related_name"] = related_name.lower()



import click

from edgy.cli.base import check as _check
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.command()
def check(env: MigrationEnv, directory: str) -> None:
    """Check if there are any new operations to migrate"""
    _check(env.app, directory)



from collections.abc import Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
    cast,
)

import sqlalchemy
from pydantic import BaseModel

from edgy.core.db.constants import CASCADE
from edgy.core.db.context_vars import CURRENT_PHASE
from edgy.core.db.fields.base import BaseForeignKey
from edgy.core.db.fields.factories import ForeignKeyFieldFactory
from edgy.core.db.fields.types import BaseFieldType
from edgy.core.db.relationships.relation import (
    SingleRelation,
    VirtualCascadeDeletionSingleRelation,
)
from edgy.exceptions import FieldDefinitionError
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType


FK_CHAR_LIMIT = 63


class BaseForeignKeyField(BaseForeignKey):
    force_cascade_deletion_relation: bool = False
    relation_has_post_delete_callback: bool = False

    def __init__(
        self,
        *,
        on_update: str,
        on_delete: str,
        related_fields: Sequence[str] = (),
        no_constraint: bool = False,
        embed_parent: Optional[tuple[str, str]] = None,
        relation_fn: Optional[Callable[..., ManyRelationProtocol]] = None,
        reverse_path_fn: Optional[Callable[[str], tuple[Any, str, str]]] = None,
        remove_referenced: bool = False,
        null: bool = False,
        **kwargs: Any,
    ) -> None:
        self.related_fields = related_fields
        self.on_update = on_update
        self.on_delete = on_delete
        self.no_constraint = no_constraint
        self.embed_parent = embed_parent
        self.relation_fn = relation_fn
        self.reverse_path_fn = reverse_path_fn
        self.remove_referenced = remove_referenced
        if remove_referenced:
            self.post_delete_callback = self._notset_post_delete_callback
        # if null:
        #     kwargs.setdefault("default", None)
        super().__init__(**kwargs, null=null)
        if self.force_cascade_deletion_relation or (
            self.on_delete == CASCADE and self.no_constraint
        ):
            self.relation_has_post_delete_callback = True

    async def _notset_post_delete_callback(self, value: Any, instance: "BaseModelType") -> None:
        value = self.expand_relationship(value)
        if value is not None:
            await value.delete(remove_referenced_call=True)

    async def pre_save_callback(
        self, value: Any, original_value: Any, force_insert: bool, instance: "BaseModelType"
    ) -> dict[str, Any]:
        target = self.target
        # value is clean result, check what is provided as kwarg
        # still use value for handling defaults
        if value is None or (isinstance(value, dict) and not value):
            value = original_value
        # e.g. default was a Model
        if isinstance(value, (target, target.proxy_model)):
            await value.save()
            return self.clean(self.name, value, for_query=False)
        elif isinstance(value, dict):
            return await self.pre_save_callback(
                target(**value), original_value=None, force_insert=force_insert, instance=instance
            )
        # don't mess around when None, we cannot save something here
        if value is None:
            return {}
        return {self.name: value}

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        if self.relation_fn is not None:
            return self.relation_fn(**kwargs)
        if self.force_cascade_deletion_relation or (
            self.on_delete == CASCADE and self.no_constraint
        ):
            relation: Any = VirtualCascadeDeletionSingleRelation
        else:
            relation = SingleRelation
        return cast(
            ManyRelationProtocol,
            relation(
                to=self.owner, to_foreign_key=self.name, embed_parent=self.embed_parent, **kwargs
            ),
        )

    def traverse_field(self, path: str) -> tuple[Any, str, str]:
        return self.target, self.reverse_name, path.removeprefix(self.name).removeprefix("__")

    def reverse_traverse_field(self, path: str) -> tuple[Any, str, str]:
        if self.reverse_path_fn:
            return self.reverse_path_fn(path)
        return self.owner, self.name, path.removeprefix(self.reverse_name).removeprefix("__")

    @cached_property
    def related_columns(self) -> dict[str, Optional[sqlalchemy.Column]]:
        target = self.target
        columns: dict[str, Optional[sqlalchemy.Column]] = {}
        if self.related_fields:
            for field_name in self.related_fields:
                if field_name in target.meta.fields:
                    for column in target.meta.field_to_columns[field_name]:
                        columns[column.key] = column
                else:
                    # placeholder for extracting column
                    columns[field_name] = None
        else:
            # try to use explicit primary keys
            if target.pknames:
                for pkname in target.pknames:
                    for column in target.meta.field_to_columns[pkname]:
                        columns[column.key] = column
            elif target.pkcolumns:
                # placeholder for extracting column
                # WARNING: this can recursively loop
                columns = {col: None for col in target.pkcolumns}
        return columns

    def expand_relationship(self, value: Any) -> Any:
        if value is None:
            return None
        target = self.target
        related_columns = self.related_columns.keys()
        if isinstance(value, (target, target.proxy_model)):
            # if all related columns are set to None
            if all(
                key in value.__dict__ and getattr(value, key) is None for key in related_columns
            ):
                return None
            return value
        if len(related_columns) == 1 and not isinstance(value, (dict, BaseModel)):
            value = {next(iter(related_columns)): value}
        elif isinstance(value, BaseModel):
            return self.expand_relationship({col: getattr(value, col) for col in related_columns})
        instance = target.proxy_model(**value)
        instance.identifying_db_fields = related_columns
        return instance

    def clean(self, name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        retdict: dict[str, Any] = {}
        column_names = self.owner.meta.field_to_column_names[name]
        assert len(column_names) >= 1
        if value is None:
            for column_name in column_names:
                retdict[column_name] = None
        elif isinstance(value, dict):
            for column_name in column_names:
                translated_name = self.from_fk_field_name(name, column_name)
                if translated_name in value:
                    retdict[column_name] = value[translated_name]
        elif isinstance(value, BaseModel):
            for column_name in column_names:
                translated_name = self.from_fk_field_name(name, column_name)
                if hasattr(value, translated_name):
                    retdict[column_name] = getattr(value, translated_name)
        elif len(column_names) == 1:
            column_name = next(iter(column_names))
            retdict[column_name] = value
        else:
            raise ValueError(f"cannot handle: {value} of type {type(value)}")
        return retdict

    def modify_input(self, name: str, kwargs: dict[str, Any]) -> None:
        phase = CURRENT_PHASE.get()
        column_names = self.get_column_names(name)
        assert len(column_names) >= 1
        if len(column_names) == 1:
            # fake default
            if phase in {"post_insert", "post_update", "load"}:
                kwargs.setdefault(name, None)
            return
        to_add = {}
        # for idempotency
        for column_name in column_names:
            if column_name in kwargs:
                to_add[self.from_fk_field_name(name, column_name)] = kwargs.pop(column_name)
        # empty
        if not to_add:
            # fake default
            if phase in {"post_insert", "post_update", "load"}:
                kwargs.setdefault(name, None)
            return
        if name in kwargs:
            # after removing the attributes return
            return
        if len(column_names) != len(to_add):
            raise ValueError("Cannot update the foreign key partially")
        kwargs[name] = to_add

    def get_fk_name(self, name: str) -> str:
        """
        Builds the fk name for the engine.

        Engines have a limitation of the foreign key being bigger than 63
        characters.

        if that happens, we need to assure it is small.
        """
        fk_name = f"fk_{self.owner.meta.tablename}_{self.target.meta.tablename}_{name}"
        if not len(fk_name) > FK_CHAR_LIMIT:
            return fk_name
        return fk_name[:FK_CHAR_LIMIT]

    def get_fkindex_name(self, name: str) -> str:
        """
        Builds the fk name for the engine.

        Engines have a limitation of the foreign key being bigger than 63
        characters.

        if that happens, we need to assure it is small.
        """
        fk_name = f"fkindex_{self.owner.meta.tablename}_{self.target.meta.tablename}_{name}"
        if not len(fk_name) > FK_CHAR_LIMIT:
            return fk_name
        return fk_name[:FK_CHAR_LIMIT]

    def get_fk_field_name(self, name: str, fieldname: str) -> str:
        if len(self.related_columns) == 1:
            return name
        return f"{name}_{fieldname}"

    def from_fk_field_name(self, name: str, fieldname: str) -> str:
        if len(self.related_columns) == 1:
            return next(iter(self.related_columns.keys()))
        return fieldname.removeprefix(f"{name}_")

    def get_columns(self, name: str) -> Sequence[sqlalchemy.Column]:
        target = self.target
        columns = []
        for column_name, related_column in self.related_columns.items():
            if related_column is None:
                related_column = target.table.columns[column_name]
            fkcolumn_name = self.get_fk_field_name(name, column_name)
            # use the related column as reference
            fkcolumn = sqlalchemy.Column(
                fkcolumn_name,
                related_column.type,
                primary_key=self.primary_key,
                autoincrement=False,
                nullable=related_column.nullable or self.null,
                unique=related_column.unique,
            )
            columns.append(fkcolumn)
        return columns

    def get_global_constraints(
        self,
        name: str,
        columns: Sequence[sqlalchemy.Column],
        schemes: Sequence[str] = (),
        no_constraint: Optional[bool] = None,
    ) -> Sequence[Union[sqlalchemy.Constraint, sqlalchemy.Index]]:
        constraints: list[Union[sqlalchemy.Constraint, sqlalchemy.Index]] = []
        no_constraint = bool(
            no_constraint
            or self.no_constraint
            # this does not work because fks are checked in metadata
            # this implies is_cross_db and is just a stronger version
            or self.owner.meta.registry is not self.target.meta.registry
        )
        if not no_constraint:
            target = self.target
            assert not target.__is_proxy_model__
            # use the last prefix as fallback
            prefix = ""
            for schema in schemes:
                prefix = f"{schema}.{target.meta.tablename}" if schema else target.meta.tablename
                if prefix in target.meta.registry.metadata.tables:
                    break
            constraints.append(
                sqlalchemy.ForeignKeyConstraint(
                    columns,
                    [
                        f"{prefix}.{self.from_fk_field_name(name, column.key)}"
                        for column in columns
                    ],
                    ondelete=self.on_delete,
                    onupdate=self.on_update,
                    name=self.get_fk_name(name),
                ),
            )
        # set for unique or if index is True
        if self.unique or self.index:
            constraints.append(
                sqlalchemy.Index(self.get_fkindex_name(name), *columns, unique=self.unique),
            )

        return constraints


class ForeignKey(ForeignKeyFieldFactory):
    field_bases = (BaseForeignKeyField,)
    field_type: Any = Any

    def __new__(  # type: ignore
        cls,
        to: Union[type["BaseModelType"], str],
        **kwargs: Any,
    ) -> BaseFieldType:
        return super().__new__(cls, to=to, **kwargs)

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        embed_parent = kwargs.get("embed_parent")
        if embed_parent and "__" in embed_parent[1]:
            raise FieldDefinitionError(
                '"embed_parent" second argument (for embedding parent) cannot contain "__".'
            )



from collections.abc import Sequence
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
    cast,
)

from edgy.core.db.constants import CASCADE
from edgy.core.db.context_vars import CURRENT_INSTANCE
from edgy.core.db.fields.base import BaseForeignKey
from edgy.core.db.fields.factories import ForeignKeyFieldFactory
from edgy.core.db.fields.foreign_keys import ForeignKey
from edgy.core.db.relationships.relation import ManyRelation
from edgy.core.terminal import Print
from edgy.core.utils.models import create_edgy_model
from edgy.exceptions import FieldDefinitionError
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType

terminal = Print()


def _removeprefixes(text: str, *prefixes: str) -> str:
    for prefix in prefixes:
        text = text.removeprefix(prefix)
    return text


class BaseManyToManyForeignKeyField(BaseForeignKey):
    is_m2m: bool = True

    def __init__(
        self,
        *,
        to_fields: Sequence[str] = (),
        to_foreign_key: str = "",
        from_fields: Sequence[str] = (),
        from_foreign_key: str = "",
        through: Union[str, type["BaseModelType"]] = "",
        embed_through: Union[str, Literal[False]] = "",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.to_fields = to_fields
        self.to_foreign_key = to_foreign_key
        self.from_fields = from_fields
        self.from_foreign_key = from_foreign_key
        self.through = through
        self.embed_through = embed_through

    @cached_property
    def embed_through_prefix(self) -> str:
        if self.embed_through is False:
            return ""
        if not self.embed_through:
            return self.name
        return f"{self.name}__{self.embed_through}"

    @cached_property
    def reverse_embed_through_prefix(self) -> str:
        if self.embed_through is False:
            return ""
        if not self.embed_through:
            return self.reverse_name
        return f"{self.reverse_name}__{self.embed_through}"

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        assert not isinstance(self.through, str), "through not initialized yet"
        return ManyRelation(
            through=self.through,
            to=self.target,
            from_foreign_key=self.from_foreign_key,
            to_foreign_key=self.to_foreign_key,
            embed_through=self.embed_through,
            **kwargs,
        )

    def get_reverse_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        assert not isinstance(self.through, str), "through not initialized yet"
        return ManyRelation(
            through=self.through,
            to=self.owner,
            reverse=True,
            from_foreign_key=self.to_foreign_key,
            to_foreign_key=self.from_foreign_key,
            embed_through=self.embed_through,
            **kwargs,
        )

    def traverse_field(self, path: str) -> tuple[Any, str, str]:
        if self.embed_through_prefix and path.startswith(self.embed_through_prefix):
            return (
                self.through,
                self.from_foreign_key,
                path.removeprefix(self.embed_through_prefix).removeprefix("__"),
            )
        return self.target, self.reverse_name, path.removeprefix(self.name).removeprefix("__")

    def reverse_traverse_field_fk(self, path: str) -> tuple[Any, str, str]:
        # used for target fk
        if self.reverse_embed_through_prefix and path.startswith(
            self.reverse_embed_through_prefix
        ):
            return (
                self.through,
                self.to_foreign_key,
                path.removeprefix(self.reverse_embed_through_prefix).removeprefix("__"),
            )
        return self.owner, self.name, path.removeprefix(self.reverse_name).removeprefix("__")

    def create_through_model(self) -> None:
        """
        Creates the default empty through model.

        Generates a middle model based on the owner of the field and the field itself and adds
        it to the main registry to make sure it generates the proper models and migrations.
        """
        from edgy.core.db.models.metaclasses import MetaInfo

        __bases__: tuple[type[BaseModelType], ...] = ()
        pknames = set()
        if self.through:
            if isinstance(self.through, str):
                assert self.owner.meta.registry is not None, "no registry found"
                self.through = self.owner.meta.registry.models[self.through]
            through = self.through
            if through.meta.abstract:
                pknames = set(through.pknames)
                __bases__ = (through,)
            else:
                if not self.from_foreign_key:
                    candidate = None
                    for field_name in through.meta.foreign_key_fields:
                        field = through.meta.fields[field_name]
                        if field.target == self.owner:
                            if candidate:
                                raise ValueError("multiple foreign keys to owner")
                            else:
                                candidate = field_name
                    if not candidate:
                        raise ValueError("no foreign key fo owner found")
                    self.from_foreign_key = candidate
                if not self.to_foreign_key:
                    candidate = None
                    for field_name in through.meta.foreign_key_fields:
                        field = through.meta.fields[field_name]
                        if field.target == self.target:
                            if candidate:
                                raise ValueError("multiple foreign keys to target")
                            else:
                                candidate = field_name
                    if not candidate:
                        raise ValueError("no foreign key fo target found")
                    self.to_foreign_key = candidate
                return
        owner_name = self.owner.__name__
        to_name = self.target.__name__
        class_name = f"{owner_name}{to_name}"
        if not self.from_foreign_key:
            self.from_foreign_key = owner_name.lower()

        if not self.to_foreign_key:
            self.to_foreign_key = to_name.lower()

        tablename = f"{owner_name.lower()}s_{to_name}s".lower()
        meta_args = {
            "tablename": tablename,
            "registry": self.owner.meta.registry,
            "multi_related": [to_name.lower()],
        }
        has_pknames = pknames and not pknames.issubset(
            {self.from_foreign_key, self.to_foreign_key}
        )
        if has_pknames:
            meta_args["unique_together"] = [(self.from_foreign_key, self.to_foreign_key)]

        new_meta: MetaInfo = MetaInfo(None, **meta_args)

        to_related_name: Union[str, Literal[False]]
        if self.related_name is False:
            to_related_name = False
        elif self.related_name:
            to_related_name = f"{self.related_name}"
            self.reverse_name = to_related_name
        else:
            if self.unique:
                to_related_name = f"{to_name.lower()}_{class_name.lower()}"
            else:
                to_related_name = f"{to_name.lower()}_{class_name.lower()}s_set"
            self.reverse_name = to_related_name

        # in any way m2m fields will have an index (either by unique_together or by their primary key constraint)

        fields = {
            f"{self.from_foreign_key}": ForeignKey(
                self.owner,
                null=True,
                on_delete=CASCADE,
                related_name=False,
                reverse_name=self.name,
                related_fields=self.from_fields,
                primary_key=not has_pknames,
                index=self.index,
            ),
            f"{self.to_foreign_key}": ForeignKey(
                self.target,
                null=True,
                on_delete=CASCADE,
                unique=self.unique,
                related_name=to_related_name,
                related_fields=self.to_fields,
                embed_parent=(self.from_foreign_key, self.embed_through or ""),
                primary_key=not has_pknames,
                index=self.index,
                relation_fn=self.get_reverse_relation,
                reverse_path_fn=self.reverse_traverse_field_fk,
            ),
        }

        # Create the through model, which adds itself to registry
        through_model = create_edgy_model(
            __name__=class_name,
            __module__=self.__module__,
            __definitions__=fields,
            __metadata__=new_meta,
            __bases__=__bases__,
        )
        self.through = through_model

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        """
        Meta field
        """
        instance = cast("BaseModelType", CURRENT_INSTANCE.get())
        if isinstance(value, ManyRelationProtocol):
            return {field_name: value}
        if instance:
            relation_instance = self.__get__(instance)
            if not isinstance(value, Sequence):
                value = [value]
            relation_instance.stage(*value)
        else:
            relation_instance = self.get_relation(refs=value)
        return {field_name: relation_instance}

    def has_default(self) -> bool:
        """Checks if the field has a default value set"""
        return False

    def get_default_values(self, field_name: str, cleaned_data: dict[str, Any]) -> Any:
        """
        Meta field
        """
        return {}

    def __get__(self, instance: "BaseModelType", owner: Any = None) -> ManyRelationProtocol:
        if instance:
            if instance.__dict__.get(self.name, None) is None:
                instance.__dict__[self.name] = self.get_relation()
            if instance.__dict__[self.name].instance is None:
                instance.__dict__[self.name].instance = instance
            return instance.__dict__[self.name]  # type: ignore
        raise ValueError("Missing instance")

    async def post_save_callback(
        self, value: ManyRelationProtocol, instance: "BaseModelType", force_insert: bool
    ) -> None:
        await value.save_related()


class ManyToManyField(ForeignKeyFieldFactory):
    field_type: Any = Any
    field_bases = (BaseManyToManyForeignKeyField,)

    def __new__(  # type: ignore
        cls,
        to: Union["BaseModelType", str],
        *,
        through: Optional["BaseModelType"] = None,
        from_fields: Sequence[str] = (),
        to_fields: Sequence[str] = (),
        **kwargs: Any,
    ) -> "BaseFieldType":
        return super().__new__(
            cls, to=to, through=through, from_fields=from_fields, to_fields=to_fields, **kwargs
        )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        embed_through = kwargs.get("embed_through")
        if embed_through and "__" in embed_through:
            raise FieldDefinitionError('"embed_through" cannot contain "__".')

        for argument in ["null", "on_delete", "on_update"]:
            if kwargs.get(argument):
                terminal.write_warning(
                    f"Declaring `{argument}` on a ManyToMany relationship has no effect."
                )
        kwargs["null"] = True
        kwargs["on_delete"] = CASCADE
        kwargs["on_update"] = CASCADE


ManyToMany = ManyToManyField



import os
from typing import TYPE_CHECKING

from databasez.testclient import DatabaseTestClient as _DatabaseTestClient

if TYPE_CHECKING:
    pass

default_test_prefix: str = "test_"
# for allowing empty
if "EDGY_TESTCLIENT_TEST_PREFIX" in os.environ:
    default_test_prefix = os.environ["EDGY_TESTCLIENT_TEST_PREFIX"]


class DatabaseTestClient(_DatabaseTestClient):
    """
    Adaption of DatabaseTestClient for edgy.

    Note: the default of lazy_setup is True here. This enables the simple Registry syntax.
    Note: the default of full_isolation is True here.

    """

    testclient_default_test_prefix = default_test_prefix

    testclient_default_lazy_setup: bool = (
        os.environ.get("EDGY_TESTCLIENT_LAZY_SETUP", "true") or ""
    ).lower() == "true"
    testclient_default_force_rollback: bool = (
        os.environ.get("EDGY_TESTCLIENT_FORCE_ROLLBACK") or ""
    ).lower() == "true"
    testclient_default_use_existing: bool = (
        os.environ.get("EDGY_TESTCLIENT_USE_EXISTING") or ""
    ).lower() == "true"
    testclient_default_drop_database: bool = (
        os.environ.get("EDGY_TESTCLIENT_DROP_DATABASE") or ""
    ).lower() == "true"
    testclient_default_full_isolation: bool = (
        os.environ.get("EDGY_TESTCLIENT_FULL_ISOLATION", "true") or ""
    ).lower() == "true"



from contextvars import ContextVar
from typing import TYPE_CHECKING, Literal, Optional, Union

if TYPE_CHECKING:
    from edgy.core.db.models.types import BaseModelType
    from edgy.core.db.querysets.base import QuerySet

CURRENT_INSTANCE: ContextVar[Optional[Union["BaseModelType", "QuerySet"]]] = ContextVar(
    "CURRENT_INSTANCE", default=None
)
CURRENT_MODEL_INSTANCE: ContextVar[Optional["BaseModelType"]] = ContextVar(
    "CURRENT_MODEL_INSTANCE", default=None
)
CURRENT_PHASE: ContextVar[str] = ContextVar("CURRENT_PHASE", default="")
EXPLICIT_SPECIFIED_VALUES: ContextVar[Optional[set[str]]] = ContextVar(
    "EXPLICIT_SPECIFIED_VALUES", default=None
)
MODEL_GETATTR_BEHAVIOR: ContextVar[Literal["passdown", "load", "coro"]] = ContextVar(
    "MODEL_GETATTR_BEHAVIOR", default="load"
)
TENANT: ContextVar[str] = ContextVar("tenant", default=None)
SCHEMA: ContextVar[str] = ContextVar("SCHEMA", default=None)
# for bw compatibility
SHEMA = SCHEMA


def get_tenant() -> Union[str, None]:
    """
    Gets the current active tenant in the context.
    """
    return TENANT.get()


def set_tenant(value: Union[str, None]) -> None:
    """
    Sets the global tenant for the context of the queries.
    When a global tenant is set the `get_context_schema` -> `SCHEMA` is ignored.
    """
    TENANT.set(value)


def get_schema() -> Union[str, None]:
    return SCHEMA.get()


def set_schema(value: Union[str, None]) -> None:
    SCHEMA.set(value)



import os
from functools import cached_property
from pathlib import Path
from typing import Union

from pydantic_settings import BaseSettings, SettingsConfigDict


class MediaSettings(BaseSettings):
    """
    All settings related to media and media root.
    """

    file_upload_temp_dir: Union[str, None] = None
    file_upload_permissions: Union[int, None] = 0o644
    file_upload_directory_permissions: Union[int, None] = None
    # don't allow overwriting the project files by default, set to media
    media_root: Union[str, os.PathLike] = Path("media/")
    media_url: str = ""

    # Storage defaults
    storages: dict[str, dict] = {
        "default": {
            "backend": "edgy.core.files.storage.filesystem.FileSystemStorage",
        },
        # "staticfiles": {
        #     "storage": "django.contrib.staticfiles.storage.StaticFilesStorage",
        # },
    }


class EdgySettings(MediaSettings):
    model_config = SettingsConfigDict(extra="allow", ignored_types=(cached_property,))
    ipython_args: list[str] = ["--no-banner"]
    ptpython_config_file: str = "~/.config/ptpython/config.py"

    # General settings
    dialects: dict[str, str] = {"postgres": "postgres", "postgresql": "postgresql"}



"""
Client to interact with Edgy models and migrations.
"""

import inspect
import sys
import typing
from functools import wraps

import click

from edgy.cli.constants import (
    APP_PARAMETER,
    EXCLUDED_COMMANDS,
    HELP_PARAMETER,
    IGNORE_COMMANDS,
)
from edgy.cli.env import MigrationEnv
from edgy.cli.operations import (
    check,
    current,
    downgrade,
    edit,
    heads,
    history,
    init,
    inspect_db,
    list_templates,
    makemigrations,
    merge,
    migrate,
    revision,
    shell,
    show,
    stamp,
)
from edgy.core.terminal import Print
from edgy.exceptions import CommandEnvironmentError

printer = Print()


class EdgyGroup(click.Group):
    """Edgy command group with extras for the commands"""

    def add_command(self, cmd: click.Command, name: typing.Optional[str] = None) -> None:
        if cmd.callback:
            cmd.callback = self.wrap_args(cmd.callback)
        return super().add_command(cmd, name)

    def wrap_args(self, func: typing.Any) -> typing.Any:
        params = inspect.signature(func).parameters

        @wraps(func)
        def wrapped(ctx: click.Context, /, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            scaffold = ctx.ensure_object(MigrationEnv)
            if "env" in params:
                kwargs["env"] = scaffold
            return func(*args, **kwargs)

        return click.pass_context(wrapped)

    def invoke(self, ctx: click.Context) -> typing.Any:
        """
        Migrations can be ignored depending of the functionality from what is being
        called.
        """
        path = ctx.params.get("path", None)

        # Process any settings
        if HELP_PARAMETER not in sys.argv and not any(
            value in sys.argv for value in EXCLUDED_COMMANDS
        ):
            try:
                migration = MigrationEnv()
                app_env = migration.load_from_env(path=path)
                ctx.obj = app_env
            except CommandEnvironmentError as e:
                if not any(value in sys.argv for value in IGNORE_COMMANDS):
                    printer.write_error(str(e))
                    sys.exit(1)
        return super().invoke(ctx)


@click.group(cls=EdgyGroup)
@click.option(
    APP_PARAMETER,
    "path",
    help="Module path to the application to generate the migrations. In a module:path format.",
)
@click.pass_context
def edgy_cli(ctx: click.Context, path: typing.Optional[str]) -> None:
    """Performs database migrations"""
    ...


edgy_cli.add_command(list_templates)
edgy_cli.add_command(init, name="init")
edgy_cli.add_command(revision, name="revision")
edgy_cli.add_command(makemigrations, name="makemigrations")
edgy_cli.add_command(edit, name="edit")
edgy_cli.add_command(merge, name="merge")
edgy_cli.add_command(migrate, name="migrate")
edgy_cli.add_command(downgrade, name="downgrade")
edgy_cli.add_command(show, name="show")
edgy_cli.add_command(history, name="history")
edgy_cli.add_command(heads, name="heads")
edgy_cli.add_command(current, name="current")
edgy_cli.add_command(stamp, name="stamp")
edgy_cli.add_command(check, name="check")
edgy_cli.add_command(shell, name="shell")
edgy_cli.add_command(inspect_db, name="inspectdb")



import sys
from functools import wraps
from typing import Any, TypeVar

from alembic.util import CommandError
from loguru import logger

T = TypeVar("T")


def catch_errors(fn: T) -> T:
    @wraps(fn)  # type: ignore
    def wrap(*args: Any, **kwargs: Any) -> T:
        try:
            fn(*args, **kwargs)  # type: ignore
        except (CommandError, RuntimeError) as exc:
            logger.error(f"Error: {str(exc)}")
            sys.exit(1)

    return wrap  # type: ignore



from typing import TYPE_CHECKING, Any, ClassVar, Optional, Union, cast

import sqlalchemy
from pydantic_core._pydantic_core import SchemaValidator as SchemaValidator

from edgy.core.utils.sync import run_sync
from edgy.exceptions import ImproperlyConfigured
from edgy.types import Undefined

if TYPE_CHECKING:
    from edgy import Registry
    from edgy.core.connection.database import Database


class ReflectedModelMixin:
    """
    Reflect on async engines is not yet supported, therefore, we need to make a sync_engine
    call.
    """

    __reflected__: ClassVar[bool] = True

    @classmethod
    def build(
        cls, schema: Optional[str] = None, metadata: Optional[sqlalchemy.MetaData] = None
    ) -> Any:
        """
        The inspect is done in an async manner and reflects the objects from the database.
        """
        registry = cls.meta.registry
        assert registry is not None, "registry is not set"
        if metadata is None:
            metadata = registry.metadata
        if cls.__using_schema__ is not Undefined:
            schema_name = cls.__using_schema__
        else:
            schema_name = schema or registry.db_schema

        tablename: str = cast("str", cls.meta.tablename)
        return run_sync(cls.reflect(cls.database, tablename, metadata, schema_name))

    @classmethod
    def fields_not_supported_by_table(
        cls, table: sqlalchemy.Table, check_type: bool = True
    ) -> set[str]:
        """Check if the model fields are a subset of the table."""
        field_names = set()
        for field_name, field in cls.meta.fields.items():
            field_has_typing_check = not field.skip_reflection_type_check and check_type
            for column in cls.meta.field_to_columns[field_name]:
                if (
                    # string not in is not supported by sqlalchemy
                    table.columns.get(column.key) is None
                    or (
                        field_has_typing_check
                        and column.type.as_generic().__class__
                        != table.columns[column.key].type.as_generic().__class__
                    )
                ):
                    field_names.add(field_name)

        return field_names

    @classmethod
    async def reflect(
        cls,
        registry: Union["Registry", "Database"],
        tablename: str,
        metadata: sqlalchemy.MetaData,
        schema: Union[str, None] = None,
    ) -> sqlalchemy.Table:
        """
        Reflect a table from the database and return its SQLAlchemy Table object.

        This method connects to the database using the provided registry, reflects
        the table with the given name and metadata, and returns the SQLAlchemy
        Table object.

        Parameters:
            registry (Registry): The registry object containing the database engine.
            tablename (str): The name of the table to reflect.
            metadata (sqlalchemy.MetaData): The SQLAlchemy MetaData object to associate with the reflected table.
            schema (Union[str, None], optional): The schema name where the table is located. Defaults to None.

        Returns:
            sqlalchemy.Table: The reflected SQLAlchemy Table object.

        Raises:
            ImproperlyConfigured: If there is an error during the reflection process.
        """

        def execute_reflection(connection: sqlalchemy.Connection) -> sqlalchemy.Table:
            """Helper function to create and reflect the table."""
            try:
                return sqlalchemy.Table(
                    tablename, metadata, schema=schema, autoload_with=connection
                )
            except Exception as e:
                raise e

        if hasattr(registry, "database"):
            registry = registry.database
        try:
            async with registry as database:
                table: sqlalchemy.Table = await database.run_sync(execute_reflection)
        except Exception as e:
            raise ImproperlyConfigured(detail=str(e)) from e
        unsupported_fields = cls.fields_not_supported_by_table(table)
        if unsupported_fields:
            raise ImproperlyConfigured(
                "Following fields have columns not matching the table specification:",
                ", ".join(unsupported_fields),
            )

        return table



from functools import cached_property
from typing import TYPE_CHECKING, Any, Union, cast

from edgy.core.db.constants import CASCADE
from edgy.core.db.context_vars import get_schema
from edgy.core.db.fields.foreign_keys import BaseForeignKeyField, ForeignKey
from edgy.core.db.relationships.relation import SingleRelation
from edgy.core.terminal import Print
from edgy.protocols.many_relationship import ManyRelationProtocol
from edgy.types import Undefined

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.types import BaseModelType

terminal = Print()


class BaseContentTypeFieldField(BaseForeignKeyField):
    async def pre_save_callback(
        self, value: Any, original_value: Any, force_insert: bool, instance: "BaseModelType"
    ) -> dict[str, Any]:
        target = self.target
        if value is None or (isinstance(value, dict) and not value):
            value = original_value
        # e.g. default was a Model
        if isinstance(value, (target, target.proxy_model)):
            value.name = self.owner.__name__
            if instance.__using_schema__ is Undefined:
                value.schema_name = get_schema()
            else:
                value.schema_name = instance.__using_schema__
        return await super().pre_save_callback(
            value, original_value, force_insert=force_insert, instance=instance
        )

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        if self.relation_fn is not None:
            return self.relation_fn(**kwargs)
        return cast(
            ManyRelationProtocol,
            SingleRelation(
                to=self.owner, to_foreign_key=self.name, embed_parent=self.embed_parent, **kwargs
            ),
        )

    def get_default_value(self) -> Any:
        default = getattr(self, "default", None)
        if callable(default):
            # WARNING: here defaults are called with the owner
            return default(self.owner)
        return default

    @cached_property
    def reverse_name(self) -> str:
        return f"reverse_{self.owner.__name__.lower()}"

    @cached_property
    def related_name(self) -> str:
        return f"reverse_{self.owner.__name__.lower()}"


class ContentTypeField(ForeignKey):
    field_bases = (BaseContentTypeFieldField,)

    def __new__(  # type: ignore
        cls,
        to: Union[type["BaseModelType"], str] = "ContentType",
        on_delete: str = CASCADE,
        no_constraint: bool = False,
        delete_orphan: bool = True,
        default: Any = lambda owner: owner.meta.registry.get_model("ContentType")(
            name=owner.__name__
        ),
        **kwargs: Any,
    ) -> "BaseFieldType":
        return super().__new__(
            cls,
            to=to,
            default=default,
            on_delete=on_delete,
            no_constraint=no_constraint,
            delete_orphan=delete_orphan,
            **kwargs,
        )

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        for argument in ["related_name", "reverse_name", "unique", "null"]:
            if kwargs.get(argument):
                terminal.write_warning(
                    f"Declaring `{argument}` on a ContentTypeField has no effect."
                )
        kwargs.pop("related_name", None)
        kwargs.pop("reverse_name", None)
        kwargs["unique"] = True
        kwargs["null"] = False
        super().validate(kwargs)



from abc import ABC, abstractmethod
from typing import Any


class BaseExtra(ABC):
    @abstractmethod
    def set_edgy_extension(self, app: Any, **kwargs: Any) -> None:
        raise NotImplementedError(
            """
            Any class implementing the extra structure must implement set_edgy_extension().

            This is a hook which must be called in `__init__` to register a config in the app.
            """
        )



import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from edgy.core.connection.database import Database


def check_db_connection(db: "Database") -> None:
    if not db.is_connected:
        # with force_rollback the effects are even worse, so fail
        if db.force_rollback:
            raise RuntimeError("db is not connected.")
        # db engine will be created and destroyed afterwards
        warnings.warn(
            "Database not connected. Executing operation is inperformant.",
            UserWarning,
            stacklevel=2,
        )



from typing import TYPE_CHECKING, Any, Optional

from pydantic_core import PydanticUndefined

if TYPE_CHECKING:
    from pydantic import ConfigDict

    from edgy.core.db.models import Model
    from edgy.core.db.models.base import EdgyBaseModel
    from edgy.core.db.models.metaclasses import MetaInfo
    from edgy.core.db.models.types import BaseModelType


def create_edgy_model(
    __name__: str,
    __module__: str,
    __definitions__: Optional[dict[str, Any]] = None,
    __metadata__: Optional["MetaInfo"] = None,
    __qualname__: Optional[str] = None,
    __config__: Optional["ConfigDict"] = None,
    __bases__: Optional[tuple[type["BaseModelType"], ...]] = None,
    __proxy__: bool = False,
    __pydantic_extra__: Any = None,
) -> type["Model"]:
    """
    Generates an `edgy.Model` with all the required definitions to generate the pydantic
    like model.
    """
    from edgy.core.db.models.model import Model

    if not __bases__:
        __bases__ = (Model,)

    qualname = __qualname__ or __name__
    core_definitions = {
        "__module__": __module__,
        "__qualname__": qualname,
        "__is_proxy_model__": __proxy__,
    }
    if not __definitions__:
        __definitions__ = {}

    core_definitions.update(**__definitions__)

    if __config__:
        core_definitions.update(**{"model_config": __config__})
    if __metadata__:
        core_definitions.update(**{"Meta": __metadata__})
    if __pydantic_extra__:
        core_definitions.update(**{"__pydantic_extra__": __pydantic_extra__})

    model: type[Model] = type(__name__, __bases__, core_definitions)
    return model


def generify_model_fields(model: type["EdgyBaseModel"]) -> dict[Any, Any]:
    """
    Makes all fields generic when a partial model is generated or used.
    This also removes any metadata for the field such as validations making
    it a clean slate to be used internally to process dynamic data and removing
    the constraints of the original model fields.
    """
    fields = {}

    # handle the nested non existing results
    for name, field in model.model_fields.items():
        field.annotation = Any  # type: ignore
        field.null = True
        # set a default to fix is_required
        if field.default is PydanticUndefined:
            field.default = None
        field.metadata = []
        fields[name] = field
    return fields



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional

from edgy.core.db.fields.file_field import FileField
from edgy.core.files import ImageFieldFile
from edgy.exceptions import FieldDefinitionError

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType


class ImageField(FileField):
    def __new__(  # type: ignore
        cls,
        # image formats without approval
        image_formats: Optional[Sequence[str]] = (),
        # extra image formats after approval
        approved_image_formats: Optional[Sequence[str]] = None,
        field_file_class: type[ImageFieldFile] = ImageFieldFile,
        **kwargs: Any,
    ) -> "BaseFieldType":
        kwargs.setdefault("with_approval", True)
        return super().__new__(
            cls,
            field_file_class=field_file_class,
            image_formats=image_formats,
            approved_image_formats=approved_image_formats,
            **kwargs,
        )

    @classmethod
    def extract_metadata(
        cls, field_obj: "BaseFieldType", field_name: str, field_file: "ImageFieldFile"
    ) -> dict[str, Any]:
        from PIL import UnidentifiedImageError

        data: dict[str, Any] = super().extract_metadata(
            field_obj, field_name=field_name, field_file=field_file
        )
        assert isinstance(field_file, ImageFieldFile)
        # here the formats are checked
        try:
            img = field_file.open_image()
            data["height"] = img.height
            data["width"] = img.width
            img.close()
        except UnidentifiedImageError:
            pass

        return data

    @classmethod
    def validate(cls, kwargs: dict[str, Any]) -> None:
        super().validate(kwargs)
        try:
            import PIL  # noqa: F401
        except ImportError:
            raise FieldDefinitionError(
                "pillow library is missing. Cannot use ImageField"
            ) from None



import uuid
import warnings
from datetime import date
from typing import Any, Optional, Union, cast
from uuid import UUID

from loguru import logger

import edgy
from edgy import settings
from edgy.core.db.models.model import Model
from edgy.core.db.models.utils import get_model
from edgy.core.utils.db import check_db_connection
from edgy.exceptions import ModelSchemaError, ObjectNotFound


class TenantMixin(edgy.Model):
    """
    Abstract table that acts as an entry-point for
    the tenants with Edgy contrib.
    """

    schema_name: str = edgy.CharField(max_length=63, unique=True, index=True)
    domain_url: str = edgy.URLField(null=True, default=settings.domain, max_length=2048)
    tenant_name: str = edgy.CharField(max_length=100, unique=True, null=False)
    tenant_uuid: UUID = edgy.UUIDField(default=uuid.uuid4, null=False)
    paid_until: date = edgy.DateField(null=True)
    on_trial: bool = edgy.BooleanField(null=True)  # type: ignore
    created_on: date = edgy.DateField(auto_now_add=True)

    # Default True, the schema will be automatically created and synched when it is saved.
    auto_create_schema: bool = getattr(settings, "auto_create_schema", True)
    """
    Set this flag to false on a parent class if you don't want the schema to be automatically
    generated.
    """
    auto_drop_schema: bool = getattr(settings, "auto_drop_schema", False)
    """
    Use with caution! Set this flag to true if you want the schema to be dropped if the
    tenant row is deleted from this table.
    """

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return f"{self.tenant_name} - {self.schema_name}"

    async def save(
        self,
        force_insert: bool = False,
        values: Union[dict[str, Any], set[str], None] = None,
        force_save: Optional[bool] = None,
    ) -> Model:
        """
        Creates a tenant record and generates a schema in the database.

        When a schema is created, then generates the tables for that same schema
        from the tenant models.
        """
        if force_save is not None:
            warnings.warn(
                "'force_save' is deprecated in favor of 'force_insert'",
                DeprecationWarning,
                stacklevel=2,
            )
            force_insert = force_save
        fields = self.extract_db_fields()
        schema_name = fields.get("schema_name", None)

        if (
            not schema_name
            or schema_name.lower() == settings.tenant_schema_default.lower()
            or schema_name == self.meta.registry.db_schema
        ):
            current_schema = (
                settings.tenant_schema_default.lower()
                if not self.meta.registry.db_schema
                else self.meta.registry.db_schema
            )
            raise ModelSchemaError(
                f"Can't update tenant outside it's own schema or the public schema. Current schema is '{current_schema}'"
            )

        tenant = await super().save(force_insert, values)
        registry = self.meta.registry
        assert registry is not None, "registry is not set"
        try:
            await registry.schema.create_schema(
                schema=tenant.schema_name,
                if_not_exists=True,
                init_tenant_models=True,
                update_cache=False,
            )
        except Exception as e:
            message = f"Rolling back... {str(e)}"
            logger.error(message)
            await self.delete()
        return tenant

    async def delete(self, force_drop: bool = False) -> None:
        """
        Validates the permissions for the schema before deleting it.
        """
        if self.schema_name == settings.tenant_schema_default:
            raise ValueError("Cannot drop public schema.")
        registry = self.meta.registry
        assert registry is not None, "registry is not set"

        await registry.schema.drop_schema(schema=self.schema_name, cascade=True, if_exists=True)
        await super().delete()


class DomainMixin(edgy.Model):
    """
    All models that store the domains must use this class
    """

    domain: str = edgy.CharField(max_length=253, unique=True, db_index=True)
    tenant: Any = edgy.ForeignKey(settings.tenant_model, index=True, related_name="domains")
    is_primary: bool = edgy.BooleanField(default=True, index=True)  # type: ignore

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return self.domain

    async def save(
        self: Any,
        force_insert: bool = False,
        values: Union[dict[str, Any], set[str], None] = None,
        force_save: Optional[bool] = None,
    ) -> Model:
        if force_save is not None:
            force_insert = force_save
        check_db_connection(self.database)
        async with self.database as database, database.transaction():
            domains = self.__class__.query.filter(tenant=self.tenant, is_primary=True).exclude(
                pk=self.pk
            )

            exists = await domains.exists()

            self.is_primary = self.is_primary or (not exists)
            if self.is_primary:
                await domains.update(is_primary=False)

            return await super().save(force_insert, values)

    async def delete(self) -> None:
        tenant = await self.tenant.load()
        if (
            tenant.schema_name.lower() == settings.tenant_schema_default.lower()
            and self.domain == settings.domain_name
        ):
            raise ValueError("Cannot drop public domain.")
        return await super().delete()


class TenantUserMixin(edgy.Model):
    """
    Mapping between user and a client (tenant).
    """

    user: Any = edgy.ForeignKey(
        settings.auth_user_model,
        null=False,
        related_name="tenant_user_users",
    )
    tenant: Any = edgy.ForeignKey(
        settings.tenant_model,
        null=False,
        related_name="tenant_users_tenant",
    )
    is_active: bool = edgy.BooleanField(default=False)  # type: ignore
    created_on: date = edgy.DateField(auto_now_add=True)

    class Meta:
        abstract = True

    @classmethod
    async def get_active_user_tenant(cls, user: edgy.Model) -> Union[edgy.Model, None]:
        """
        Obtains the active user tenant.
        """
        try:
            tenant = await get_model(
                registry=cls.meta.registry, model_name=cls.__name__
            ).query.get(user=user, is_active=True)
            await tenant.tenant.load()

        except ObjectNotFound:
            return None
        return cast(edgy.Model, tenant.tenant)

    def __str__(self) -> str:
        return f"User: {self.user.pk}, Tenant: {self.tenant}"

    async def save(self, *args: Any, **kwargs: Any) -> edgy.Model:
        await super().save(*args, **kwargs)
        if self.is_active:
            await (
                get_model(registry=self.meta.registry, model_name=self.__class__.__name__)
                .query.filter(is_active=True, user=self.user)
                .exclude(pk=self.pk)
                .update(is_active=False)
            )
        return self



import functools
import inspect
import typing
from collections.abc import Sequence
from typing import Any, Callable, Optional, TypeVar

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

_T = TypeVar("_T")


def is_async_callable(obj: typing.Any) -> bool:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return inspect.iscoroutinefunction(obj) or (
        callable(obj) and inspect.iscoroutinefunction(obj.__call__)
    )


class AyncLifespanContextManager:
    """
    Manages and handles the on_startup and on_shutdown events
    in an Esmerald way.

    This is not the same as the on_startup and on_shutdown
    from Starlette. Those are now deprecated and will be removed
    in the version 1.0 of Starlette.

    This aims to provide a similar functionality but by generating
    a lifespan event based on the values from the on_startup and on_shutdown
    lists.
    """

    def __init__(
        self,
        on_shutdown: Optional[Sequence[Callable[..., Any]]] = None,
        on_startup: Optional[Sequence[Callable[..., Any]]] = None,
    ) -> None:
        self.on_startup = [] if on_startup is None else list(on_startup)
        self.on_shutdown = [] if on_shutdown is None else list(on_shutdown)

    def __call__(self: _T, app: object) -> _T:
        return self

    async def __aenter__(self) -> None:
        """Runs the functions on startup"""
        for handler in self.on_startup:
            if is_async_callable(handler):
                await handler()
            else:
                handler()

    async def __aexit__(self, scope: Scope, receive: Receive, send: Send, **kwargs: Any) -> None:
        """Runs the functions on shutdown"""
        for handler in self.on_shutdown:
            if is_async_callable(handler):
                await handler()
            else:
                handler()


def handle_lifespan_events(
    on_startup: Optional[Sequence[Callable]] = None,
    on_shutdown: Optional[Sequence[Callable]] = None,
    lifespan: Optional[Any] = None,
) -> Any:
    """Handles with the lifespan events in the new Starlette format of lifespan.
    This adds a mask that keeps the old `on_startup` and `on_shutdown` events variable
    declaration for legacy and comprehension purposes and build the async context manager
    for the lifespan.
    """
    if on_startup or on_shutdown:
        return AyncLifespanContextManager(on_startup=on_startup, on_shutdown=on_shutdown)
    elif lifespan:
        return lifespan
    return None



import os
import pathlib
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, TypeVar, Union

from edgy.core.files.base import ContentFile, File
from edgy.exceptions import SuspiciousFileOperation
from edgy.utils.path import get_random_string, get_valid_filename, validate_file_name

_arg_val = TypeVar("_arg_val")
_arg_setting = TypeVar("_arg_setting")


class Storage(ABC):
    """
    The base storage that should be used
    to implement/override the default
    storage of Edgy.
    """

    # automatically set by handler
    name: str = ""

    # private helper
    @staticmethod
    def value_or_setting(value: _arg_val, setting: _arg_setting) -> Union[_arg_val, _arg_setting]:
        return setting if value is None else value

    @abstractmethod
    def _open(self, name: str, mode: str) -> Any: ...

    def open(self, name: str, mode: Union[str, None] = None) -> Any:
        if mode is None:
            mode = "rb"
        return self._open(name, mode)

    @abstractmethod
    def _save(self, content: "File", name: str = "") -> None: ...

    def save(self, content: Any, name: str = "") -> None:
        """
        Save new content to the file specified by name. The content should be
        a proper File object or any Python file-like object, ready to be read
        from the beginning.
        """
        if not name:
            name = content.name

        name = self.sanitize_name(name)

        if isinstance(content, str):
            content = ContentFile(content.encode("utf8"), name)
        elif isinstance(content, bytes):
            content = ContentFile(content, name)
        elif not hasattr(content, "chunks"):
            content = File(content, name)

        self._save(content, name)

    @abstractmethod
    def reserve_name(self, name: str) -> bool: ...

    @abstractmethod
    def unreserve_name(self, name: str) -> bool: ...

    def sanitize_name(self, name: str) -> str:
        """
        Return a filename, based on the provided filename, that's suitable for
        use in the target storage system.
        """
        validate_file_name(name, allow_relative_path=True)
        return get_valid_filename(name)

    def get_alternative_name(self, file_root: str, file_ext: str) -> str:
        """
        Return an alternative filename, by adding an underscore and a random 7
        character alphanumeric string (before the file extension, if one
        exists) to the filename.
        """
        return f"{file_root}_{get_random_string(7)}{file_ext}"

    def get_available_name(
        self,
        name: str,
        max_length: Union[int, None] = None,
        overwrite: bool = False,
        multi_process_safe: Union[bool] = None,
    ) -> str:
        """
        Return a filename that's free on the target storage system and
        available for new content to be written to.

        Args:
            name (str): The original filename.
            max_length (int, optional): The maximum length of the filename. Defaults to None.
            overwrite (bool, optional): Don't generate alternative names.

        Returns:
            str: The available filename. Need to call unreserve_name afterwards.

        Raises:
            SuspiciousFileOperation: If a path traversal attempt is detected or an available filename cannot be found.
        """
        if multi_process_safe is None:
            multi_process_safe = not overwrite
        name = str(name).replace("\\", "/")
        dir_name, file_name = os.path.split(name)

        if ".." in pathlib.PurePath(dir_name).parts:
            raise SuspiciousFileOperation(f"Detected path traversal attempt in '{dir_name}'")

        validate_file_name(file_name)
        file_root, file_ext = os.path.splitext(file_name)
        if multi_process_safe:
            # Ddd hexified process pid to prevent collisions among processes.
            # Adding proccess pid in front because it is most important to prevent classhes
            # among multiple processes.
            # Note: despite there is a thread id it is not always available and threads
            # often shortlived, so use a lock instead (see FilesystemStorage).
            file_root = f"{os.getpid() or 0:x}_{file_root}"

        while (max_length and len(name) > max_length) or (
            not self.reserve_name(name) and not overwrite
        ):
            if not overwrite:
                name = os.path.join(dir_name, self.get_alternative_name(file_root, file_ext))
            else:
                name = os.path.join(dir_name, f"{file_root}{file_ext}")

            if max_length is not None:
                truncation = len(name) - max_length
                if truncation > 0:
                    file_root = file_root[:-truncation]
                    if not file_root:
                        raise SuspiciousFileOperation(
                            f'Storage can not find an available filename for "{name}". '
                            "Please make sure that the corresponding file field "
                            'allows sufficient "max_length".'
                        )
                    if not overwrite:
                        name = os.path.join(
                            dir_name, self.get_alternative_name(file_root, file_ext)
                        )
                    else:
                        name = os.path.join(dir_name, f"{file_root}{file_ext}")

        return name

    def path(self, name: str) -> str:
        """
        Return a local filesystem path where the file can be retrieved using
        Python's built-in open() function. Storage systems that can't be
        accessed using open() should *not* implement this method.
        """
        raise NotImplementedError("This backend doesn't support absolute paths.")

    @abstractmethod
    def delete(self, name: str) -> None:
        """
        Delete the specified file from the storage system.
        """

    @abstractmethod
    def exists(self, name: str) -> bool:
        """
        Return True if a file referenced by the given name already exists in the
        storage system, or False if the name is available for a new file.
        """

    @abstractmethod
    def listdir(self, path: str) -> tuple[list[str], list[str]]:
        """
        List the contents of the specified path. Return a 2-tuple of lists:
        the first item being directories, the second item being files.
        """

    @abstractmethod
    def size(self, name: str) -> int:
        """
        Return the total size, in bytes, of the file specified by name.
        """

    def url(self, name: str) -> str:
        """
        Return an absolute URL where the file's contents can be accessed
        directly by a web browser.
        """
        raise NotImplementedError("This backend doesn't support 'url'.")

    def get_accessed_time(self, name: str) -> datetime:
        """
        Return the last accessed time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("This backend doesn't support 'accessed_time'.")

    def get_created_time(self, name: str) -> datetime:
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("This backend doesn't support 'created_time'.")

    def get_modified_time(self, name: str) -> datetime:
        """
        Return the last modified time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError("This backend doesn't support 'modified_time'.")



import click

from edgy.cli.base import edit as _edit
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.command()
@click.argument("revision", default="head")
def edit(env: MigrationEnv, directory: str, revision: str) -> None:
    """Edit a revision file"""
    _edit(env.app, directory, revision)



import warnings
from typing import TYPE_CHECKING, Union

import sqlalchemy
from loguru import logger

from edgy.core.terminal import Terminal
from edgy.exceptions import ModelSchemaError

terminal = Terminal()

if TYPE_CHECKING:
    from edgy import Model, Registry
    from edgy.core.db.models.types import BaseModelType


def table_schema(model_class: type["Model"], schema: str) -> sqlalchemy.Table:
    warnings.warn(
        "'table_schema' has been deprecated, use '<model>.table_schema' instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return model_class.table_schema(schema)


async def create_tables(
    registry: "Registry", models: dict[str, type["BaseModelType"]], schema: str
) -> None:
    """
    Creates the table models for a specific schema just generated.

    Iterates through the tenant models and creates them in the schema.
    """

    for model in models.values():
        table = model.table_schema(schema)

        logger.info(f"Creating table '{model.meta.tablename}' for schema: '{schema}'")
        try:
            async with registry.database as database:
                await database.run_sync(table.create)
        except Exception as e:
            logger.error(str(e))
            ...


async def create_schema(
    registry: "Registry",
    schema_name: str,
    models: Union[dict[str, type["BaseModelType"]], None] = None,
    if_not_exists: bool = False,
    should_create_tables: bool = False,
) -> None:
    """
    Creates a schema in a given registry.

    This function creates a new schema for a tenant in the provided registry.
    It optionally checks if the schema already exists and creates tables within
    the schema if specified.

    Parameters:
        registry (Registry): The registry object where the schema will be created.
        schema_name (str): The name of the schema to be created.
        models (dict[str, Model], optional): When models is provided, it will use them to generate
                                             those same models in a given schema or else
                                             it will default to the `registry.models`.
        if_not_exists (bool, optional): If True, the schema will be created only if it does not already exist.
                                        Defaults to False.
        should_create_tables (bool, optional): If True, tables will be created within the new schema. Defaults to False.

    Raises:
        ModelSchemaError: If the schema name is the same as the default schema name.
    """
    default_schema_name = registry.schema.get_default_schema() or "public"
    if schema_name.lower() == default_schema_name.lower():
        raise ModelSchemaError(
            f"Cannot create a schema with the same name as the default: '{schema_name}'."
        )

    # Create the new schema, optionally checking if it already exists
    await registry.schema.create_schema(schema_name, if_not_exists=if_not_exists)

    # Optionally create tables within the new schema
    if should_create_tables:
        terminal.write_info(f"Creating the tables in schema: {schema_name}")
        schema_models = models or registry.models
        await create_tables(registry, schema_models, schema_name)



from typing import TYPE_CHECKING, ClassVar, cast

import edgy

from .metaclasses import ContentTypeMeta

if TYPE_CHECKING:
    from edgy.core.db.querysets.base import QuerySet


class ContentType(edgy.Model, metaclass=ContentTypeMeta):
    no_constraint: ClassVar[bool] = False

    class Meta:
        abstract = True

    # NOTE: model_ is a private namespace of pydantic
    # model names shouldn't be so long, maybe a check would be appropriate
    name: str = edgy.fields.CharField(max_length=100, default="", index=True)
    # set also the schema for tenancy support
    schema_name: str = edgy.CharField(max_length=63, null=True, index=True)
    # can be a hash or similar. For checking collisions cross domain
    collision_key: str = edgy.fields.CharField(max_length=255, null=True, unique=True)

    async def get_instance(self) -> edgy.Model:
        reverse_name = f"reverse_{self.name.lower()}"
        return (
            await cast("QuerySet", getattr(self, reverse_name))
            .using(schema=self.schema_name)
            .get()
        )

    async def delete(
        self, skip_post_delete_hooks: bool = False, remove_referenced_call: bool = False
    ) -> None:
        reverse_name = f"reverse_{self.name.lower()}"
        query = cast("QuerySet", getattr(self, reverse_name))
        await super().delete(skip_post_delete_hooks=skip_post_delete_hooks)
        if not remove_referenced_call and self.no_constraint:
            await query.using(schema=self.schema_name).delete()



import copy
import operator
from typing import Any, Callable, TypeVar

empty = object()
RT = TypeVar("RT")


def new_method_proxy(func: Callable[..., RT]) -> Callable[..., RT]:
    def inner(self, *args: Any) -> RT:  # type: ignore
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    return inner


class LazyObject:  # pragma: no cover
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class.
    By subclassing, you have the opportunity to intercept and alter the
    instantiation. If you don't need to do that, use SimpleLazyObject.
    """

    # Avoid infinite recursion when tracing __init__ (#19456).
    _wrapped = None

    def __init__(self) -> None:
        # Note: if a subclass overrides __init__(), it will likely need to
        # override __copy__() and __deepcopy__() as well.
        self._wrapped = empty

    def __getattribute__(self, name: str) -> Any:
        if name == "_wrapped":
            # Avoid recursion when getting wrapped object.
            return super().__getattribute__(name)
        value: Any = super().__getattribute__(name)
        # If attribute is a proxy method, raise an AttributeError to call
        # __getattr__() and use the wrapped object method.
        if not getattr(value, "_mask_wrapped", True):
            raise AttributeError
        return value

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_wrapped":
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name: str) -> None:
        if name == "_wrapped":
            raise TypeError("can't delete _wrapped.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self) -> Any:
        """
        Must be implemented by subclasses to initialize the wrapped object.
        """
        raise NotImplementedError("subclasses of LazyObject must provide a _setup() method")

    # Because we have messed with __class__ below, we confuse pickle as to what
    # class we are pickling. We're going to have to initialize the wrapped
    # object to successfully pickle it, so we might as well just pickle the
    # wrapped object since they're supposed to act the same way.
    #
    # Unfortunately, if we try to simply act like the wrapped object, the ruse
    # will break down when pickle gets our id(). Thus we end up with pickle
    # thinking, in effect, that we are a distinct object from the wrapped
    # object, but with the same __dict__. This can cause problems (see #25389).
    #
    # So instead, we define our own __reduce__ method and custom unpickler. We
    # pickle the wrapped object as the unpickler's argument, so that pickle
    # will pickle it normally, and then the unpickler simply returns its
    # argument.
    def __reduce__(self) -> Any:
        if self._wrapped is empty:
            self._setup()
        return (unpickle_lazyobject, (self._wrapped,))

    def __copy__(self) -> Any:
        if self._wrapped is empty:
            # If uninitialized, copy the wrapper. Use type(self), not
            # self.__class__, because the latter is proxied.
            return type(self)()
        else:
            # If initialized, return a copy of the wrapped object.
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo: Any) -> Any:
        if self._wrapped is empty:
            # We have to use type(self), not self.__class__, because the
            # latter is proxied.
            result = type(self)()
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __bytes__: Any = new_method_proxy(bytes)
    __str__: Any = new_method_proxy(str)
    __bool__: Any = new_method_proxy(bool)

    # Introspection support
    __dir__: Any = new_method_proxy(dir)

    # Need to pretend to be the wrapped class, for the sake of objects that
    # care about this (especially in equality tests)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))  # type: ignore
    __eq__: Any = new_method_proxy(operator.eq)
    __lt__: Any = new_method_proxy(operator.lt)
    __gt__: Any = new_method_proxy(operator.gt)
    __ne__: Any = new_method_proxy(operator.ne)
    __hash__: Any = new_method_proxy(hash)

    # List/Tuple/Dictionary methods support
    __getitem__: Any = new_method_proxy(operator.getitem)
    __setitem__: Any = new_method_proxy(operator.setitem)
    __delitem__: Any = new_method_proxy(operator.delitem)
    __iter__: Any = new_method_proxy(iter)
    __len__: Any = new_method_proxy(len)
    __contains__: Any = new_method_proxy(operator.contains)


def unpickle_lazyobject(wrapped: Any) -> Any:  # pragma: no cover
    """
    Used to unpickle lazy objects. Just return its argument, which will be the
    wrapped object.
    """
    return wrapped


class SimpleLazyObject(LazyObject):  # pragma: no cover
    """
    A lazy object initialized from any function.
    Designed for compound objects of unknown type. For builtins or objects of
    known type, use django.utils.functional.lazy.
    """

    def __init__(self, func: Callable[..., RT]) -> None:
        """
        Pass in a callable that returns the object to be wrapped.
        If copies are made of the resulting SimpleLazyObject, which can happen
        in various circumstances within Django, then you must ensure that the
        callable can be safely run more than once and will return the same
        value.
        """
        self.__dict__["_setupfunc"] = func
        super().__init__()

    def _setup(self) -> Any:
        self._wrapped = self._setupfunc()

    # Return a meaningful representation of the lazy object for debugging
    # without evaluating the wrapped object.
    def __repr__(self) -> str:
        repr_attr = self._setupfunc if self._wrapped is empty else self._wrapped
        return f"<{type(self).__name__}: {repr_attr!r}>"

    def __copy__(self) -> Any:
        if self._wrapped is empty:
            # If uninitialized, copy the wrapper. Use SimpleLazyObject, not
            # self.__class__, because the latter is proxied.
            return SimpleLazyObject(self._setupfunc)
        else:
            # If initialized, return a copy of the wrapped object.
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo: Any) -> Any:
        if self._wrapped is empty:
            # We have to use SimpleLazyObject, not self.__class__, because the
            # latter is proxied.
            result = SimpleLazyObject(self._setupfunc)
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __add__ = new_method_proxy(operator.add)

    @new_method_proxy
    def __radd__(self, other: Any) -> Any:
        return other + self



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import sqlalchemy
from pydantic import BaseModel, Field

from edgy.types import Undefined

if TYPE_CHECKING:
    from edgy.core.connection import Registry
    from edgy.core.db.fields.factories import FieldFactory
    from edgy.core.db.models.metaclasses import MetaInfo
    from edgy.core.db.models.types import BaseModelType


class _ColumnDefinition:
    # column specific
    primary_key: bool = False
    autoincrement: bool = False
    index: bool = False
    unique: bool = False
    default: Any = None
    comment: Optional[str] = None
    server_default: Optional[Any] = None
    server_onupdate: Optional[Any] = None


# for preventing shadowing warnings
class ColumnDefinition(_ColumnDefinition):
    null: bool = False
    column_name: Optional[str] = None
    column_type: Any = None
    constraints: Sequence[sqlalchemy.Constraint] = ()


class ColumnDefinitionModel(
    _ColumnDefinition, BaseModel, extra="ignore", arbitrary_types_allowed=True
):
    null: bool = Field(serialization_alias="nullable", default=False)
    column_name: Optional[str] = Field(exclude=True, default=None)
    column_type: Any = Field(exclude=True, default=None)
    constraints: Sequence[sqlalchemy.Constraint] = Field(exclude=True, default=())


class BaseFieldDefinitions:
    read_only: bool = False
    inject_default_on_partial_update: bool = False
    inherit: bool = True
    skip_absorption_check: bool = False
    skip_reflection_type_check: bool = False
    field_type: Any = Any
    factory: Optional[FieldFactory] = None

    __original_type__: Any = None
    name: str = ""
    secret: bool = False
    exclude: bool = False
    owner: Any = None
    default: Any = Undefined
    explicit_none: bool = False
    server_default: Any = None

    # column specific
    primary_key: bool = False
    autoincrement: bool = False
    null: bool = False
    index: bool = False
    unique: bool = False


class BaseFieldType(BaseFieldDefinitions, ABC):
    @abstractmethod
    def is_required(self) -> bool:
        """Check if the argument is required.

        Returns:
            `True` if the argument is required, `False` otherwise.
        """

    @abstractmethod
    def has_default(self) -> bool:
        """Checks if the field has a default value set"""

    @abstractmethod
    def get_columns(self, field_name: str) -> Sequence[sqlalchemy.Column]:
        """
        Returns the columns of the field being declared.
        """

    def operator_to_clause(
        self, field_name: str, operator: str, table: sqlalchemy.Table, value: Any
    ) -> Any:
        """
        Analyzes the operator and build a clause from it.
        Needs to be able to handle "exact".
        """
        raise NotImplementedError()

    def clean(self, field_name: str, value: Any, for_query: bool = False) -> dict[str, Any]:
        """
        Validates a value and transform it into columns which can be used for querying and saving.

        Args:
            field_name: the field name (can be different from name)
            value: the field value
        Kwargs:
            for_query: Is used for querying. Should have all columns used for querying set.
                       The columns used can differ especially for multi column fields.
        """
        return {}

    def to_model(
        self,
        field_name: str,
        value: Any,
    ) -> dict[str, Any]:
        """
        Inverse of clean. Transforms column(s) to a field for a pydantic model (EdgyBaseModel).
        Validation happens later.

        Args:
            field_name: the field name (can be different from name)
            value: the field value
        Kwargs:
            phase: the phase (set, creation, ...)

        """
        return {field_name: value}

    def get_global_constraints(
        self, name: str, columns: Sequence[sqlalchemy.Column], schemes: Sequence[str] = ()
    ) -> Sequence[Union[sqlalchemy.Constraint, sqlalchemy.Index]]:
        """Return global constraints and indexes.
        Useful for multicolumn fields
        """
        return []

    def get_embedded_fields(
        self, field_name: str, fields: dict[str, BaseFieldType]
    ) -> dict[str, BaseFieldType]:
        """
        Define extra fields on the fly. Often no owner is available yet.

        Args:
            field_name: the field name (can be different from name)
            fields: the existing fields

        Note: the returned fields are changed after return, so you should
              return new fields or copies. Also set the owner of the field to them before returning
        """
        return {}

    @abstractmethod
    def get_default_values(self, field_name: str, cleaned_data: dict[str, Any]) -> Any:
        """
        Define for each field/column a default. Non-private multicolumn fields should
        always check if the default was already applied.

        Args:
            field_name: the field name (can be different from name)
            cleaned_data: currently validated data. Useful to check if the default was already applied.

        """

    @abstractmethod
    def embed_field(
        self,
        prefix: str,
        new_fieldname: str,
        owner: Optional[BaseModelType] = None,
        parent: Optional[BaseFieldType] = None,
    ) -> Optional[BaseFieldType]:
        """
        Embed this field or return None to prevent embedding.
        Must return a copy with name and owner set when not returning None.
        Args:
            prefix: the prefix
            new_field_name: the new field name
            owner: The new owner if available
            parent: The parent field which embeds the field if available
        """

    # helpers

    @property
    def registry(self) -> Registry:
        warnings.warn(
            "registry attribute of field is deprecated, use 'owner.meta.registry' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.owner.meta.registry  # type: ignore

    def get_column_names(self, name: str = "") -> frozenset[str]:
        if name:
            return cast("MetaInfo", self.owner.meta).field_to_column_names[name]
        return cast("MetaInfo", self.owner.meta).field_to_column_names[self.name]



import typing
from inspect import isclass
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from typing_extensions import get_origin

import edgy
from edgy.core.db.context_vars import CURRENT_PHASE
from edgy.core.db.fields.base import BaseField
from edgy.core.db.fields.factories import ForeignKeyFieldFactory
from edgy.exceptions import ModelReferenceError

if TYPE_CHECKING:
    from edgy.core.db.fields.types import BaseFieldType
    from edgy.core.db.models.model_reference import ModelRef
    from edgy.core.db.models.types import BaseModelType


class RefForeignKey(ForeignKeyFieldFactory, list):
    field_type = list
    field_bases = (BaseField,)

    @classmethod
    def modify_input(
        cls,
        field_obj: "BaseFieldType",
        name: str,
        kwargs: dict[str, Any],
        original_fn: Any = None,
    ) -> None:
        phase = CURRENT_PHASE.get()
        # we are empty
        if name not in kwargs and phase == "init_db":
            kwargs[name] = []

    def embed_field(
        self,
        prefix: str,
        new_fieldname: str,
        owner: Optional[type["BaseModelType"]] = None,
        parent: Optional["BaseFieldType"] = None,
    ) -> Optional["BaseFieldType"]:
        return None

    @classmethod
    def is_class_and_subclass(cls, value: typing.Any, _type: typing.Any) -> bool:
        original = get_origin(value)
        if not original and not isclass(value):
            return False

        try:
            if original:
                return original and issubclass(original, _type)
            return issubclass(value, _type)
        except TypeError:
            return False

    def __new__(cls, to: "ModelRef", null: bool = False) -> "BaseFieldType":  # type: ignore
        if not cls.is_class_and_subclass(to, edgy.ModelRef):
            raise ModelReferenceError(
                detail="A model reference must be an object of type ModelRef"
            )
        if not getattr(to, "__related_name__", None):
            raise ModelReferenceError(
                "'__related_name__' must be declared when subclassing ModelRef."
            )

        return super().__new__(cls, to=to, exclude=True, null=null)

    @classmethod
    async def post_save_callback(
        cls,
        obj: "BaseFieldType",
        value: Optional[list],
        instance: "BaseModelType",
        force_insert: bool,
        original_fn: Any = None,
    ) -> None:
        if not value:
            return

        relation_field = instance.meta.fields[obj.to.__related_name__]
        extra_params = {}
        try:
            # m2m or foreign key
            target_model_class = relation_field.target
        except AttributeError:
            # reverse m2m or foreign key
            target_model_class = relation_field.related_from
        if not relation_field.is_m2m:
            # sometimes the foreign key is required, so set it already
            extra_params[relation_field.foreign_key.name] = instance
        relation = getattr(instance, obj.to.__related_name__)
        while value:
            instance_or_dict: Union[dict, ModelRef] = value.pop()
            if isinstance(instance_or_dict, dict):
                instance_or_dict = obj.to(**instance_or_dict)
            model = target_model_class(
                **cast("ModelRef", instance_or_dict).model_dump(exclude={"__related_name__"}),
                **extra_params,
            )
            # we are in a relationship field
            await relation.add(model)
